/*
 * Segment 4 Datenübernahme und Aufbereitung, SWE 4.5 Abfrage Puffer
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.abfrpuffer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectCollection;
import de.bsvrz.dua.abfrpuffer.Station.KeinKonfigurationsObjektException;
import de.bsvrz.dua.abfrpuffer.SystemDatenEndgeraet.SystemDeStatus;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.bitctrl.dua.DUAKonstanten;
import de.bsvrz.sys.funclib.bitctrl.dua.ObjektWecker;
import de.bsvrz.sys.funclib.bitctrl.dua.dfs.typen.SWETyp;
import de.bsvrz.sys.funclib.bitctrl.dua.schnittstellen.IObjektWeckerListener;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;
import de.bsvrz.sys.funclib.dataIdentificationSettings.EndOfSettingsListener;
import de.bsvrz.sys.funclib.dataIdentificationSettings.SettingsManager;
import de.bsvrz.sys.funclib.dataIdentificationSettings.UpdateListener;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

/**
 * Sorgt dafuer, dass die Puffer der DatenEndgeraaete (DeLve und DeAxl)
 * regelmaesig abgefragt werden.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 *
 */
public class AbfragePufferdaten implements StandardApplication, ClientReceiverInterface, ClientSenderInterface {

	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die PID des verwendeten Abfragemoduls.
	 */
	private static String abfrageModul;
	/**
	 * Fuer die Auslesung der eigenschaften des Archivs.
	 */
	private static SettingsManager archivSettingsManager;

	/**
	 * Verbindung zum Datenverteiler.
	 */
	private ClientDavInterface dav;
	/**
	 * Diese Klasse versendet Betriebsmeldungen.
	 */
	private final MessageSender nachrichtenSender = MessageSender.getInstance();
	/**
	 * Parameter dem man erkennen soll SteuerModule und Konfigurationsbereiche
	 * die verwaltet werden sollen.
	 */
	protected static final String P_SM = "-sm";
	/**
	 * Ob alle DEs online sein muessen, damit man eine Pufferabfrage machen
	 * kann.
	 */
	protected static final String P_ALLE_ONLINE = "-alle-online";
	/**
	 * Ob alle DEs online sein muessen, damit man eine Pufferabfrage machen
	 * kann.
	 */
	protected static final String P_ABFRAGE_MODUL = "-abfragemodul";
	/**
	 * Mutex fuer dem AbfrageModul Einstellungen.
	 */
	private final Object abfrageModulEinstellungenMutex = new Object();
	/**
	 * Streambasierte Anfrage um Infos aus dem Archiv auslesen.
	 */
	private StreamBasierteArchivAnfrage sbAnfrage;
	/**
	 * Datenbeschreibung für Globale Fehlermeludngen.
	 */
	private static DataDescription ddGloFehler;
	/**
	 * Datenbeschreibung für Abfragemodul Parametrierung.
	 */
	private static DataDescription ddAbfrageModul;
	/**
	 * SystemObjekte der SteuerModule die, die als Parameter angegeben wurden.
	 */
	private final Collection<SystemObject> sysObjSteuerModule = new LinkedList<>();
	/**
	 * SystemObjekte der konfigurationsbereichen die, die als Parameter
	 * angegeben wurden.
	 */
	private final Collection<ConfigurationArea> sysObjKonfBereiche = new LinkedList<>();
	/**
	 * PIDs der SteuerModule und Konfigurationsbereiche, die als Parameter
	 * angegeben wurden.
	 */
	private String[] paramKonfigurationsObjekte;
	/**
	 * Mutex fuer dem settingsManager und die Listener Threads zu verwalten.
	 */
	private final Object datenZugriffMutex = new Object();
	/**
	 * Liste der bearbeitenden Stationen, enthaelt alle benoetigte
	 * informationen.
	 */
	private final List<Station> stationen = new ArrayList<>();
	/**
	 * Verwaltet die Lve.
	 */
	private AbstraktDeVerwaltung lveDeVerwaltung;
	/**
	 * Verwaltet die Axl Abrufe.
	 */
	private AbstraktDeVerwaltung axlDeVerwaltung;
	/**
	 * Stellt aggregierte LZD zur Verfuegung fuer andere Module.
	 */
	private AggregatorVerwaltung lzdAggregation;
	/**
	 * Ob wir fuer dem Empfang von GloFehler angemeldet sind.
	 */
	private boolean erstmalAngemeldet;
	/**
	 * Ob alle DEs online sein muessen um eine Abfrage zu machen.
	 */
	private boolean alleMuessenOnlineSein = true;
	/**
	 * Intervall der periodischen Axl Abfragen.
	 */
	private long abfrageIntervallAxl;
	/**
	 * Enthaelt das neue Wert, wenn sich der AbfrageIntervall geaendert.
	 */
	private long abfrageIntervallAxlNeu;
	/**
	 * Zeitpunkt der naechsten Axl Abfrage.
	 */
	private long naechsteAbfrageAxl;
	/**
	 * Zeitpunkt der naechsten Abfrage, wenn sich der Offset oder Abrufintervall
	 * geaendert.
	 */
	private long naechsteAbfrageAxlNeu;
	/**
	 * Intervall der periodischen Lve Abfragen.
	 */
	private long abfrageIntervallLve;
	/**
	 * Enthaelt das neue Wert, wenn sich der AbfrageIntervall geaendert.
	 */
	private long abfrageIntervallLveNeu;
	/**
	 * Zeitpunkt der naechsten Lve Abfrage.
	 */
	private long naechsteAbfrageLve;
	/**
	 * Zeitpunkt der naechsten Abfrage, wenn sich der Offset oder Abrufintervall
	 * geaendert.
	 */
	private long naechsteAbfrageLveNeu;
	/**
	 * Der Wecker ermoeglicht regelmaesige priodische Abrufe, paralell fuer
	 * jedem Steuermodul.
	 *
	 */
	private final ObjektWecker wecker = new ObjektWecker();
	/**
	 * Listener wird beim Start der Axl Abfrage vom Wecker gerufen.
	 */
	private IObjektWeckerListener axlAbfragen;

	/**
	 * Listener wird beim Start der LveAbrfagen vom Wecker gerufen.
	 */
	private IObjektWeckerListener lveAbfragen;

	/**
	 * Listener wird vom Wecker beim Start der Aggregation gerufen.
	 */
	private IObjektWeckerListener aggregationStart;

	/**
	 * Die Aggregation muss ein bischen auf dem Archiv warten, damit das Archiv
	 * sicher alle Daten bearbeitet hat.
	 */
	private final long aggregationVerspaetung = 5 * AbfrPufferKonstanten.MINUTE_IN_MS;

	/**
	 * Wenn Zwei Zeitstempel fuer Abrufe innerhalb dieses Intervalls sind, dann
	 * koennen wir sie als gleich betrachten, vermeidet 2 Abrufe in der selben
	 * Zeit.
	 */
	protected static final long ZEIT_TOLERANZ = 500;

	@Override
	public void parseArguments(final ArgumentList argumente) throws Exception {

		Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
			@Override
			public void uncaughtException(final Thread t, final Throwable e) {
				AbfragePufferdaten.LOGGER.error("Applikation wird wegen" + " unerwartetem Fehler beendet", e);
				e.printStackTrace();
				Runtime.getRuntime().exit(-1);
			}
		});

		if (!argumente.hasArgument(AbfragePufferdaten.P_SM)) {
			AbfragePufferdaten.printVerwendung();
			throw new Exception("Keine Konfigurationsobjekte angegeben.");
		}

		String argParameter = argumente.fetchArgument(AbfragePufferdaten.P_SM).asString();
		paramKonfigurationsObjekte = argParameter.split(",");

		if (argumente.hasArgument(AbfragePufferdaten.P_ALLE_ONLINE)) {
			argParameter = argumente.fetchArgument(AbfragePufferdaten.P_ALLE_ONLINE).asString();
			if (argParameter.isEmpty()) {
				alleMuessenOnlineSein = false;
			} else {
				alleMuessenOnlineSein = Boolean.parseBoolean(argParameter);
			}
		}

		if (argumente.hasArgument(AbfragePufferdaten.P_ABFRAGE_MODUL)) {
			argParameter = argumente.fetchArgument(AbfragePufferdaten.P_ABFRAGE_MODUL).asString();
			if (argParameter.isEmpty()) {
				AbfragePufferdaten.abfrageModul = null;
			} else {
				AbfragePufferdaten.abfrageModul = argParameter;
			}
		}

		if (argumente.getArgumentStrings().length == 0) {
			AbfragePufferdaten.printVerwendung();
			System.exit(-1);
		}
	}

	/**
	 * Anwendungsparameter.
	 */
	public static void printVerwendung() {
		System.out.println("SWE Abfrage Pufferdaten (DUA)\nUnterstuetzte Parameter: ");
		System.out.println("\t" + AbfragePufferdaten.P_SM
				+ "=obj{,obj} obj ist vom Typ typ.steuerModul oder typ.konfigurationsBereich");
		System.out.println("\t" + AbfragePufferdaten.P_ALLE_ONLINE
				+ "=true|false bestimmt, ob alle Datenendgeraete (DE) eines Steuermoduls fehlerfrei arbeiten muessen, um eine Abfrage zu generieren");
		System.out.println("\t" + AbfragePufferdaten.P_ABFRAGE_MODUL
				+ "=pid eines Objektes vom Typ typ.abfrageModul, durch das diese Applikation zur Laufzeit parametriert werden kann");
	}

	@Override
	public void initialize(final ClientDavInterface connection) throws Exception {

		dav = connection;

		for (final String element : paramKonfigurationsObjekte) {
			try {
				SystemObject so = dav.getDataModel().getObject(element);

				if (so == null) {
					final long id = Long.parseLong(element);
					so = dav.getDataModel().getObject(id);
					if ((so instanceof SystemObjectCollection) && "menge.steuerModul".equals(so.getType().getPid())) {
						final SystemObjectCollection collection = (SystemObjectCollection) so;
						final List<SystemObject> liste = collection.getElements();

						for (final SystemObject obj : liste) {
							if (obj.isValid()) {
								sysObjSteuerModule.add(obj);
							}
						}
					} else {
						AbfragePufferdaten.LOGGER.error("ID " + element + " ist keine Steuermodul(menge)");
					}
				} else if (so.getType().getPid().equals(AbfrPufferKonstanten.TYP_KONFIGURATIONSBEREICH)) {
					sysObjKonfBereiche.add((ConfigurationArea) so);
				} else if (so.getType().getPid().equals(AbfrPufferKonstanten.TYP_STEUERMODUL)) {
					sysObjSteuerModule.add(so);
					stationen.add(new Station(so, dav, datenZugriffMutex));
				} else if ((so instanceof SystemObjectCollection)
						&& "menge.steuerModul".equals(so.getType().getPid())) {
					final SystemObjectCollection collection = (SystemObjectCollection) so;
					final List<SystemObject> liste = collection.getElements();

					for (final SystemObject obj : liste) {
						if (obj.isValid()) {
							sysObjSteuerModule.add(obj);
							stationen.add(new Station(so, dav, datenZugriffMutex));
						}
					}
				} else {
					AbfragePufferdaten.LOGGER.error("Pid " + element + " ist vom Typ " + so.getType().getPid()
							+ " statt Konfigurationsbereich oder Steuermodul");
				}

			} catch (final KeinKonfigurationsObjektException e) {
				AbfragePufferdaten.LOGGER.error(e.getMessage());
			}
		}
		Station.setAlleMuessenOnlineSein(alleMuessenOnlineSein);
		AbfragePufferdaten.LOGGER.info("Abfrage fuer folgende Steuermodule:" + sysObjSteuerModule
				+ "\nund/bzw. diese Konfigurationsbereiche:" + sysObjKonfBereiche);

		for (final SystemObject steuerModul : dav.getDataModel().getType(AbfrPufferKonstanten.TYP_STEUERMODUL)
				.getElements()) {
			for (final SystemObject kb : sysObjKonfBereiche) {
				if (steuerModul.getConfigurationArea().equals(kb)) {
					sysObjSteuerModule.add(steuerModul);
					stationen.add(new Station(steuerModul, dav, datenZugriffMutex));
					break;
				}
			}
		}

		String smAusKbListe = "";
		for (final SystemObject sm : sysObjSteuerModule) {
			smAusKbListe += sm.toString() + "\n";
		}
		AbfragePufferdaten.LOGGER
		.info("In gegebene Konfigurationsbereichen wurden folgende SteuerModule gefunden:" + smAusKbListe);

		// Archiv einstellungen
		final DataDescription archivDatenBeschreibung = new DataDescription(
				dav.getDataModel().getAttributeGroup(AbfrPufferKonstanten.ATG_ARCHIV),
				dav.getDataModel().getAspect(AbfrPufferKonstanten.ASP_PARAMETERSOLL));

		AbfragePufferdaten.archivSettingsManager = new SettingsManager(dav,
				new DataIdentification(dav.getDataModel().getConfigurationAuthority(), archivDatenBeschreibung));

		// Callback-Methode: Wird vom DAV aufgerufen, wenn eine an/abzumeldende
		// DID identifiziert wurde

		final UpdateListener updateSettingsLstnr = new UpdateListener() {
			@Override
			public void update(final DataIdentification did, final Data alteDatei, final Data neueDatei) {

				synchronized (datenZugriffMutex) {

					if ((alteDatei == null) && (neueDatei != null)) { // Neu
						// parametriert
						lveDeVerwaltung.bearbeiteNeueEinstellung(did, neueDatei);
						axlDeVerwaltung.bearbeiteNeueEinstellung(did, neueDatei);
					} else if ((alteDatei != null) && (neueDatei == null)) { // Nicht
						// mehr
						// parametriert
						lveDeVerwaltung.bearbeiteGeloeschteEinstellung(did, alteDatei);
						axlDeVerwaltung.bearbeiteGeloeschteEinstellung(did, alteDatei);
					} else if ((alteDatei != null) && (neueDatei != null)) { // Umparametriert
						lveDeVerwaltung.bearbeiteVeraenderteEinstellung(did, alteDatei, neueDatei);
						axlDeVerwaltung.bearbeiteVeraenderteEinstellung(did, alteDatei, neueDatei);
					}
				}
			}
		};
		AbfragePufferdaten.archivSettingsManager.addUpdateListener(updateSettingsLstnr);

		// Callback-Methode: Wird vom DAV aufgerufen, alle An/Abzumeldungen
		// beendet sind
		final EndOfSettingsListener eofSettingsLstnr = new EndOfSettingsListener() {
			@Override
			public void inform() {
				synchronized (datenZugriffMutex) {
					if (!erstmalAngemeldet) {
						anmeldeGloFehler();
						lveDeVerwaltung.ersteAnmeldung();
						axlDeVerwaltung.ersteAnmeldung();
						erstmalAngemeldet = true;
					}
					lveDeVerwaltung.anmeldung();
					axlDeVerwaltung.anmeldung();
				}
			}
		};
		AbfragePufferdaten.archivSettingsManager.addEndOfSettingsListener(eofSettingsLstnr);

		// fuer Stream Anfragen aus dem Archiv
		sbAnfrage = new StreamBasierteArchivAnfrage(dav);

		initLveAxlVerwaltung();

		lzdAggregation.addAggregator(new AggregatorFahrzeugAnzahl());
		lzdAggregation.addAggregator(new AggregatorGeschwindigkeit());
		lzdAggregation.addAggregator(new AggregatorDTV());
		lzdAggregation.addAggregator(new AggregatorVxx(15));
		lzdAggregation.addAggregator(new AggregatorVxx(50));
		lzdAggregation.addAggregator(new AggregatorVxx(85));
		lzdAggregation.addAggregator(new AggregatorSpizenStunde());
		lzdAggregation.addAggregator(new AggregatorSpizenTag());
		lzdAggregation.initialisiere(dav, currentTimeMillis());

		AbfragePufferdaten.archivSettingsManager.start();

		axlAbfragen = new IObjektWeckerListener() {
			@Override
			public void alarm() {
				axlAlarm();
			}
		};
		lveAbfragen = new IObjektWeckerListener() {
			@Override
			public void alarm() {
				lveAlarm();
			}
		};

		aggregationStart = new IObjektWeckerListener() {
			@Override
			public void alarm() {
				aggregationStart();
			}
		};
		anmeldeAbfrageModulParametrierung();
		generiereRegelmaesigeAxlAbfragen();
		generiereRegelmaesigeLveAbfragen();
	}

	/**
	 * Erfragt die PID des assoziierten Abfragemoduls.
	 *
	 * @return die PID des assoziierten Abfragemoduls bzw. <code>null</code>,
	 *         wenn keines angegeben wurde
	 */
	public static final String getAbfrageModul() {
		return AbfragePufferdaten.abfrageModul;
	}

	/**
	 * Wegen Testzwecken wird es ueberschrieben.
	 */
	public void initLveAxlVerwaltung() {
		lveDeVerwaltung = new LveDeVerwaltung(this, stationen);
		axlDeVerwaltung = new AxlDeVerwaltung(this, stationen);
		lzdAggregation = new AggregatorVerwaltung();
	}

	/**
	 * Erzeugt eine DatenBeschreibung mit gegebenen Parameter als Strings.
	 *
	 * @param atg
	 *            Atributgruppe
	 * @param asp
	 *            Aspekt
	 * @param var
	 *            Simulation Variant
	 * @return DataDescription mit gegebene parameter
	 */
	public DataDescription erzeugeDatenBeschreibung(final String atg, final String asp, final short var) {
		return new DataDescription(dav.getDataModel().getAttributeGroup(atg), dav.getDataModel().getAspect(asp), var);
	}

	/**
	 * Meldet sich online fuer die AbfrageModul Parametrierung an.
	 */
	public void anmeldeAbfrageModulParametrierung() {
		AbfragePufferdaten.ddAbfrageModul = erzeugeDatenBeschreibung(AbfrPufferKonstanten.ATG_ABFRAGEMODUL,
				AbfrPufferKonstanten.ASP_PARAMETERSOLL, DataDescription.NO_SIMULATION_VARIANT_SET);
		// LinkedList<SystemObjectType> types = new
		// LinkedList<SystemObjectType>();
		// types.add(dav.getDataModel().getType(AbfrPufferKonstanten.TYP_ABFRAGEMODUL));
		// HashSet<ConfigurationArea> konfBereich = new
		// HashSet<ConfigurationArea>();
		//
		//
		// konfBereich.addAll(this.sysObjKonfBereiche);
		// for(SystemObject so : sysObjSteuerModule) {
		// konfBereich.add(so.getConfigurationArea());
		// }
		//
		// Collection<SystemObject> abfrageModule =
		// dav.getDataModel().getObjects(konfBereich, types,
		// ObjectTimeSpecification.valid());
		// if(abfrageModule.size() != 1) {
		// LOGGER.error("Anzahl der AbfrageModul Instanzen in der eingegebenen
		// Kb: " + abfrageModule.size() + " nur eine Instanz soll existieren");
		// }

		boolean anmeldungAufModul = false;
		if (AbfragePufferdaten.getAbfrageModul() != null) {
			final SystemObject modul = dav.getDataModel().getObject(AbfragePufferdaten.getAbfrageModul());
			if (modul != null) {
				anmeldungAufModul = true;
				dav.subscribeReceiver(this, modul, AbfragePufferdaten.ddAbfrageModul, ReceiveOptions.normal(),
						ReceiverRole.receiver());
			} else {
				AbfragePufferdaten.LOGGER.warning("Uebergebenes Abfragemodul konnte nicht identifiziert werden: "
						+ AbfragePufferdaten.getAbfrageModul());
			}
		} else {
			AbfragePufferdaten.LOGGER.warning("Kein Abfragemodul uebergeben");
		}

		if (!anmeldungAufModul) {
			AbfragePufferdaten.LOGGER.info("Verwende Standardparameter");
			final long zyklus = 1, offset = 5;

			synchronized (abfrageModulEinstellungenMutex) {
				abfrageIntervallLveNeu = zyklus * AbfrPufferKonstanten.STUNDE_IN_MS;
				// System nicht virtuell, weil wir es dem Wecker weitergeben
				// Weil wir den Virtuellen Tagesbeginn brauchen, convertieren
				// wir es
				naechsteAbfrageLveNeu = convertAbsoluteVirtuelleZeit(
						AggregatorVerwaltung.getTagesBeginnVonZeitStempel(currentTimeMillis()));

				while (naechsteAbfrageLveNeu < (System.currentTimeMillis() + AbfragePufferdaten.ZEIT_TOLERANZ)) {
					naechsteAbfrageLveNeu += convertZeitIntervall(abfrageIntervallLveNeu);
				}
				naechsteAbfrageLveNeu += convertZeitIntervall(offset * AbfrPufferKonstanten.MINUTE_IN_MS);

				abfrageIntervallAxlNeu = zyklus * AbfrPufferKonstanten.STUNDE_IN_MS;
				// System nicht virtuell, weil wir es dem Wecker weitergeben
				// Weil wir den Virtuellen Tagesbeginn brauchen, convertieren
				// wir es
				naechsteAbfrageAxlNeu = convertAbsoluteVirtuelleZeit(
						AggregatorVerwaltung.getTagesBeginnVonZeitStempel(currentTimeMillis()));

				while (naechsteAbfrageAxlNeu < (System.currentTimeMillis() + AbfragePufferdaten.ZEIT_TOLERANZ)) {
					naechsteAbfrageAxlNeu += convertZeitIntervall(abfrageIntervallLveNeu);
				}
				naechsteAbfrageAxlNeu += convertZeitIntervall(offset * AbfrPufferKonstanten.MINUTE_IN_MS);

				abfrageModulEinstellungenMutex.notify();
			}
		}
	}

	/**
	 * Anmeldung fuer Empfang von Globalen FehlerMeldungen.
	 */
	public void anmeldeGloFehler() {
		if (erstmalAngemeldet) {
			return;
		}
		AbfragePufferdaten.ddGloFehler = erzeugeDatenBeschreibung(AbfrPufferKonstanten.ATG_TLS_GLODEFEHLER,
				AbfrPufferKonstanten.ASP_TLS_ANTWORT, DataDescription.NO_SIMULATION_VARIANT_SET);

		final Collection<SystemObject> allDEs = Station.getAlleDEs(stationen);
		dav.subscribeReceiver(this, allDEs, AbfragePufferdaten.ddGloFehler, ReceiveOptions.normal(),
				ReceiverRole.receiver());
		erstmalAngemeldet = true;
	}

	/**
	 * Liefert dem Nachrichtensender.
	 *
	 * @return Nachrichtensender
	 */
	public MessageSender getnachrichtenSender() {
		return nachrichtenSender;
	}

	/**
	 * Meldet sich von GloFehler Meldungen ab.
	 */
	public void abmeldeGloFehler() {
		if (!erstmalAngemeldet) {
			return;
		}
		final Collection<SystemObject> allDEs = Station.getAlleDEs(stationen);
		dav.unsubscribeReceiver(this, allDEs, AbfragePufferdaten.ddGloFehler);
		erstmalAngemeldet = false;
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		// System.out.println(object + ", " + dataDescription + ", " + state);
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		// true wenn dataRequest implementiert ist ...
		return false;
	}

	@Override
	public void update(final ResultData[] results) {
		synchronized (datenZugriffMutex) {
			for (final ResultData resDatei : results) {

				final String atgPid = resDatei.getDataDescription().getAttributeGroup().getPid();
				if (atgPid.equals(DUAKonstanten.ATG_LZD)) {
					lveDeVerwaltung.aktualisiereDeZeitStemepel(resDatei);
					lzdAggregation.update(resDatei);
				} else if (atgPid.equals(AbfrPufferKonstanten.ATG_TLS_LVEBETRIEBSPARAM)) {
					lveDeVerwaltung.setBetriebsParameter(resDatei);
				} else if (atgPid.equals(AbfrPufferKonstanten.ATG_TLS_LVENEGATIVE_QUITTUNG)) {
					lveDeVerwaltung.bearbeiteNegativeQuittung(resDatei);
				} else if (atgPid.equals(AbfrPufferKonstanten.ATG_TLS_AXLERGEBNIS_MELDUNG)) {
					axlDeVerwaltung.aktualisiereDeZeitStemepel(resDatei);
				} else if (atgPid.equals(AbfrPufferKonstanten.ATG_TLS_AXLBETRIEBSPARAM)) {
					axlDeVerwaltung.setBetriebsParameter(resDatei);
				} else if (atgPid.equals(AbfrPufferKonstanten.ATG_TLS_AXLNEGATIVE_QUITTUNG)) {
					axlDeVerwaltung.bearbeiteNegativeQuittung(resDatei);
				} else if (atgPid.equals(AbfrPufferKonstanten.ATG_TLS_GLODEFEHLER)) {
					setGloFehler(resDatei);
				} else if (atgPid.equals(AbfrPufferKonstanten.ATG_ABFRAGEMODUL)) {
					setAbrufPeriode(resDatei);
				}
			}
		}
	}

	/**
	 * Macht eine StreamAbfrage.
	 *
	 * @param bearbeiter
	 *            Der Objekt, der die Antwort bearbeiten wird, muss
	 *            SBArchivAnfrageBearbeiter implementieren
	 * @param objekte
	 *            Die Objekte, an deren Daten wir nachfragen wollen
	 * @param datenBeschreibung
	 *            Die DatenBeschreibung fuer die Daten
	 * @param relative
	 *            true, wenn die Abfrage relativ ist
	 * @param intervalBeginnOderAnzahl
	 *            Beginn des abfragenden Intervalls oder Anzahl der Datensaetze
	 *            bei relativen Abfragen
	 * @param intervalEnde
	 *            Ende des abfrafenden Intervalls
	 */
	public void streamAbfrage(final SBArchivAnfrageBearbeiter bearbeiter, final Collection<SystemObject> objekte,
			final DataDescription datenBeschreibung, final boolean relative, final long intervalBeginnOderAnzahl,
			final long intervalEnde) {

		sbAnfrage.generiereAnfrage(objekte, datenBeschreibung, relative, intervalBeginnOderAnzahl, intervalEnde);
		if (sbAnfrage.istErfolgreich()) {
			sbAnfrage.bearbeiteAntwortStream(bearbeiter);
		} else {
			AbfragePufferdaten.LOGGER.error("Archivanfrage nicht erfolgreich");
			nachrichtenSender.sendMessage(MessageType.SYSTEM_DOMAIN, MessageGrade.FATAL,
					"Archivanfrage nicht erfolgreich");
		}
	}

	/**
	 * Stelt den Status des Objektes aus einer GloFehlermeldung fest.
	 *
	 * @param datei
	 *            Datei mit atg.tlsGloFehler, die empfangen wurde
	 */
	public void setGloFehler(final ResultData datei) {
		final SystemObject de = datei.getObject();
		final Data daten = datei.getData();

		final DatenEndgeraet datenEndgeraet = Station.getDatenEndgeraet(de.getPid());
		final Station st = datenEndgeraet.getStation();

		boolean istOnline = false;

		if (daten == null) {
			istOnline = false;
		} else {
			final int status = daten.getUnscaledValue(AbfrPufferKonstanten.ATR_DE_FEHLERSTATUS).intValue();
			istOnline = AbfrPufferKonstanten.convertStatusZumBoolean(status);
		}

		final boolean warLveOnline = st.istLveOnline();
		final boolean warAxlOnline = st.istAxlOnline();
		datenEndgeraet.setOnline(istOnline, datei.getDataTime());
		final boolean istLveOnline = st.istLveOnline();
		final boolean istAxlOnline = st.istAxlOnline();

		if (warLveOnline != istLveOnline) {
			lveDeVerwaltung.aktualisiereStatus(st, "GloDEFehler-Status aktualisiert");
		}
		if (warAxlOnline != istAxlOnline) {
			axlDeVerwaltung.aktualisiereStatus(st, "GloDEFehler-Status aktualisiert");
		}
	}

	/**
	 * Stellt die Abrufperiode fest, wird bei Aenderung der Parametrierung
	 * gerufen.
	 *
	 * @param datei
	 *            Datei mit atg.abfrageModul, die empfangen wurde
	 */
	public void setAbrufPeriode(final ResultData datei) {
		final Data daten = datei.getData();
		long zyklus, offset;

		if (daten == null) {
			return;
		}

		synchronized (abfrageModulEinstellungenMutex) {
			zyklus = daten.getUnscaledValue(AbfrPufferKonstanten.ATR_LVEABFRAGEZYKLUS).longValue();
			offset = daten.getUnscaledValue(AbfrPufferKonstanten.ATR_LVEABFRAGEOFFSET).longValue();

			if (zyklus == 0) {
				AbfragePufferdaten.LOGGER.error("Lve Abfragezyklus ist 0");
				System.exit(-1);
			}
			offset = offset % 60;

			abfrageIntervallLveNeu = zyklus * AbfrPufferKonstanten.STUNDE_IN_MS;
			// System nicht virtuell, weil wir es dem Wecker weitergeben
			// Weil wir den Virtuellen Tagesbeginn brauchen, convertieren wir es
			naechsteAbfrageLveNeu = convertAbsoluteVirtuelleZeit(
					AggregatorVerwaltung.getTagesBeginnVonZeitStempel(currentTimeMillis()));

			while (naechsteAbfrageLveNeu < (System.currentTimeMillis() + AbfragePufferdaten.ZEIT_TOLERANZ)) {
				naechsteAbfrageLveNeu += convertZeitIntervall(abfrageIntervallLveNeu);
			}
			naechsteAbfrageLveNeu += convertZeitIntervall(offset * AbfrPufferKonstanten.MINUTE_IN_MS);

			zyklus = daten.getUnscaledValue(AbfrPufferKonstanten.ATR_AXLABFRAGEZYKLUS).longValue();
			offset = daten.getUnscaledValue(AbfrPufferKonstanten.ATR_AXLABFRAGEOFFSET).longValue();

			if (zyklus == 0) {
				AbfragePufferdaten.LOGGER.error("Axl Abfragezyklus ist 0");
				System.exit(-1);
			}
			offset = offset % 60;

			abfrageIntervallAxlNeu = zyklus * AbfrPufferKonstanten.STUNDE_IN_MS;
			// System nicht virtuell, weil wir es dem Wecker weitergeben
			// Weil wir den Virtuellen Tagesbeginn brauchen, convertieren wir es
			naechsteAbfrageAxlNeu = convertAbsoluteVirtuelleZeit(
					AggregatorVerwaltung.getTagesBeginnVonZeitStempel(currentTimeMillis()));

			while (naechsteAbfrageAxlNeu < (System.currentTimeMillis() + AbfragePufferdaten.ZEIT_TOLERANZ)) {
				naechsteAbfrageAxlNeu += convertZeitIntervall(abfrageIntervallLveNeu);
			}
			naechsteAbfrageAxlNeu += convertZeitIntervall(offset * AbfrPufferKonstanten.MINUTE_IN_MS);

			abfrageModulEinstellungenMutex.notify();
		}
	}

	/**
	 * Generiert eine regelmaessige Axl Abfrage.
	 */
	void generiereRegelmaesigeAxlAbfragen() {

		synchronized (abfrageModulEinstellungenMutex) {
			if ((naechsteAbfrageAxlNeu == 0) && (abfrageIntervallAxlNeu == 0) && (naechsteAbfrageAxl == 0)
					&& (abfrageIntervallAxl == 0)) {
				try {
					abfrageModulEinstellungenMutex.wait();
				} catch (final Exception e) {
					AbfragePufferdaten.LOGGER.error(e.getMessage());
					e.printStackTrace();
					System.exit(-1);
				}
			}
			// Im fall die Abfrageintervalle oder Offset hat sich geaendert
			if ((naechsteAbfrageAxlNeu != 0) && (abfrageIntervallAxlNeu != 0)) {
				naechsteAbfrageAxl = naechsteAbfrageAxlNeu;
				abfrageIntervallAxl = abfrageIntervallAxlNeu;
				naechsteAbfrageAxlNeu = 0;
				abfrageIntervallAxlNeu = 0;
			}

			// System nicht virtuell, weil wir es dem Wecker weitergeben
			while (naechsteAbfrageAxl <= (System.currentTimeMillis() + AbfragePufferdaten.ZEIT_TOLERANZ)) {
				naechsteAbfrageAxl += convertZeitIntervall(abfrageIntervallAxl);
			}

			wecker.setWecker(axlAbfragen, naechsteAbfrageAxl);
		}
	}

	/**
	 * Stellt dem Wecker fest duer die naechste regelmaesige Abfrage.
	 */
	void generiereRegelmaesigeLveAbfragen() {
		synchronized (abfrageModulEinstellungenMutex) {
			if ((naechsteAbfrageLveNeu == 0) && (abfrageIntervallLveNeu == 0) && (naechsteAbfrageLve == 0)
					&& (abfrageIntervallLve == 0)) {
				try {
					abfrageModulEinstellungenMutex.wait();
				} catch (final Exception e) {
					AbfragePufferdaten.LOGGER.error(e.getMessage());
					e.printStackTrace();
					System.exit(-1);
				}
			}
			if ((naechsteAbfrageLveNeu != 0) && (abfrageIntervallLveNeu != 0)) {
				naechsteAbfrageLve = naechsteAbfrageLveNeu;
				abfrageIntervallLve = abfrageIntervallLveNeu;
				naechsteAbfrageLveNeu = 0;
				abfrageIntervallLveNeu = 0;
			}

			while (naechsteAbfrageLve <= (System.currentTimeMillis() + AbfragePufferdaten.ZEIT_TOLERANZ)) {
				naechsteAbfrageLve += convertZeitIntervall(abfrageIntervallLve);
			}

			wecker.setWecker(lveAbfragen, naechsteAbfrageLve);
		}
	}

	/**
	 * Startet eine regelmaesige Axl Abfrage.
	 */
	public void axlAlarm() {
		synchronized (datenZugriffMutex) {
			for (final Station st : stationen) {
				if ((st.getSystemAxlDe() != null)
						&& (st.getSystemAxlDe().getStatus() == SystemDeStatus.DE_STAT_ABFRAGE_MOEGLICH)) {
					st.resetAxlNegativeQuittungen();
					axlDeVerwaltung.generiereAbfrage(st);
				}
			}
		}
		generiereRegelmaesigeAxlAbfragen();
	}

	/**
	 * Startet eine regelmaesige Lve Abfrage.
	 */
	public void lveAlarm() {
		synchronized (datenZugriffMutex) {
			for (final Station st : stationen) {
				if ((st.getSystemLveDe() != null)
						&& (st.getSystemLveDe().getStatus() == SystemDeStatus.DE_STAT_ABFRAGE_MOEGLICH)) {
					st.resetLveNegativeQuittungen();
					lveDeVerwaltung.generiereAbfrage(st);
				}
			}
		}
		generiereRegelmaesigeLveAbfragen();
	}

	/**
	 * Wird zum Start der Applikation aufgerufen.
	 *
	 * @param args
	 *            Aufrufargumente der Applikation
	 */
	public static void main(final String[] args) {
		StandardApplicationRunner.run(new AbfragePufferdaten(), args);
	}

	/**
	 * Typ der SWE.
	 *
	 * @return liefert den Typ der SWE
	 */
	public SWETyp getSWETyp() {
		return SWETyp.SWE_ABFRAGE_PUFFERDATEN;
	}

	/**
	 * Erfragt eine Instanz des DAVs.
	 *
	 * @return DAV
	 */
	public ClientDavInterface getDatenVerteiler() {
		return dav;
	}

	/**
	 * Stellt dem Wecker fuer die naechste Aggregation fest.
	 */
	public void aggregiere() {
		// System nicht virtuell, weil wir es dem Wecker weitergeben
		final long zeit = System.currentTimeMillis() + convertZeitIntervall(aggregationVerspaetung);
		wecker.setWecker(aggregationStart, zeit);
	}

	/**
	 * Startet die Aggregation von LZD.
	 */
	public void aggregationStart() {
		if (lzdAggregation == null) {
			return;
		}
		try {

			Thread.sleep(convertZeitIntervall(aggregationVerspaetung));

			synchronized (datenZugriffMutex) {
				for (final Station st : stationen) {
					final SystemDatenEndgeraet sde = st.getSystemLveDe();
					if ((sde != null)
							&& (sde.getStatus() == SystemDatenEndgeraet.SystemDeStatus.DE_STAT_ABFRAGE_LAEUFT)) {
						return;
					}
				}
				lzdAggregation.trigger();
			}
			return;
		} catch (final InterruptedException e) {
			AbfragePufferdaten.LOGGER.fine(e.getLocalizedMessage());
		}
	}

	/**
	 * Erfargt nach dem Abrufintervall fuer Axl Daten.
	 *
	 * @return Zeitintervall
	 */
	public long getAbrufIntervallAxl() {
		return abfrageIntervallAxl;
	}

	/**
	 * Erfraget nach den Abrufintervall fuer Lve Daten.
	 *
	 * @return Zeitintervall
	 */
	public long getAbrufIntervallLve() {
		return abfrageIntervallLve;
	}

	/**
	 * Wegen Test Zwecken wird es in vererbten Klassen ueberdeckt.
	 *
	 * @return Zeit
	 */
	public long currentTimeMillis() {
		return System.currentTimeMillis();
	}

	/**
	 * Wegen Test Zwecken wird es in vererbten Klassen ueberdeckt.
	 *
	 * @param zeitIntervall
	 *            Virtueller ZeitIntervall
	 * @return Reeller ZeitIntervall
	 */
	public long convertZeitIntervall(final long zeitIntervall) {
		return zeitIntervall;
	}

	/**
	 * Wegen Test Zwecken wird es in vererbten Klassen ueberdeckt.
	 *
	 * @param intervall
	 *            Virtueller Timeout Intervall
	 * @return Reeller Intervall
	 */
	public long convertTimeoutIntervall(final long intervall) {
		return intervall;
	}

	/**
	 * Wegen Test Zwecken wird es in vererbten Klassen ueberdeckt.
	 *
	 * @param zeit
	 *            Echter Zeitpunkt
	 * @return Virtueller Zeitpunkt
	 */
	public long convertAbsoluteEchtZeit(final long zeit) {
		return zeit;
	}

	/**
	 * Wegen Test Zwecken wird es in vererbten Klassen ueberdeckt.
	 *
	 * @param zeit
	 *            Virtueller Zeitpunkt
	 * @return Echter Zeitpunkt
	 */
	public long convertAbsoluteVirtuelleZeit(final long zeit) {
		return zeit;
	}

	/**
	 * liefert den aktiven Settingsmanager für den Archivdatensatz.
	 *
	 * @return den Settingsmanager
	 */
	public static SettingsManager getArchivSettingsManager() {
		return AbfragePufferdaten.archivSettingsManager;
	}

	protected AbstraktDeVerwaltung getLveDeVerwaltung() {
		return lveDeVerwaltung;
	}

	protected AbstraktDeVerwaltung getAxlDeVerwaltung() {
		return axlDeVerwaltung;
	}

	protected void setLveDeVerwaltung(final AbstraktDeVerwaltung lveDeVerwaltung) {
		this.lveDeVerwaltung = lveDeVerwaltung;
	}

	protected void setAxlDeVerwaltung(final AbstraktDeVerwaltung axlDeVerwaltung) {
		this.axlDeVerwaltung = axlDeVerwaltung;
	}

	protected List<Station> getStationen() {
		return stationen;
	}

	protected long getNaechsteAbfrageLveNeu() {
		return naechsteAbfrageLveNeu;
	}

	protected long getNaechsteAbfrageLve() {
		return naechsteAbfrageLve;
	}

	protected long getNaechsteAbfrageAxlNeu() {
		return naechsteAbfrageAxlNeu;
	}

	protected long getNaechsteAbfrageAxl() {
		return naechsteAbfrageAxl;
	}

	protected void setLzdAggregation(final AggregatorVerwaltung lzdAggregation) {
		this.lzdAggregation = lzdAggregation;
	}

	protected ClientDavInterface getDav() {
		return dav;
	}
}
